const l = {
  /**
   * Throws an error indicating an attempt to access the value of a Nothing instance.
   * @throws {Error} Throws an error with a specific message.
   */
  FromJustIsNothing() {
    throw new Error("Maybe: Attempted to access value of Nothing");
  }
};
class o {
  /**
   * Constructs a Maybe instance.
   * @param hasValue - Indicates whether the instance has a value.
   * @param value - The value contained in the Maybe instance (if any).
   */
  constructor(e, t) {
    this.hasValue = e, this.value = t;
  }
  /**
   * Checks if the Maybe instance is Nothing (has no value).
   * @returns {boolean} True if the instance is Nothing, otherwise false.
   */
  IsNothing() {
    return !this.hasValue;
  }
  /**
   * Checks if the Maybe instance is Just (has a value).
   * @returns {boolean} True if the instance is Just, otherwise false.
   */
  IsJust() {
    return this.hasValue;
  }
  /**
   * Creates a Maybe instance representing Nothing.
   * @returns {Maybe<T>} A Maybe instance with no value.
   */
  static Nothing() {
    return new o(!1);
  }
  /**
   * Creates a Maybe instance representing Just with a given value.
   * @param value - The value to be contained in the Maybe instance.
   * @returns {Maybe<T>} A Maybe instance with the provided value.
   */
  static Just(e) {
    return new o(!0, e);
  }
  /**
   * Retrieves the value of a Just instance or throws an error if it's Nothing.
   * Alias for FromJust method.
   * @returns {T} The value contained in the Just instance.
   * @throws {Error} If the instance is Nothing.
   */
  ToChecked() {
    return this.FromJust();
  }
  /**
   * Checks if the Maybe instance is Nothing and throws an error if it is.
   * @throws {Error} If the instance is Nothing.
   */
  Check() {
    this.IsNothing() && l.FromJustIsNothing();
  }
  /**
   * Retrieves the value of a Just instance and sets it in the provided output object.
   * @param out - An object to hold the value if the instance is Just.
   * @returns {boolean} True if the instance is Just, otherwise false.
   */
  To(e) {
    return this.IsJust() ? (e.value = this.value, e.hasValue = !0, !0) : !1;
  }
  /**
   * Retrieves the value of a Just instance or throws an error if it's Nothing.
   * @returns {T} The value contained in the Just instance.
   * @throws {Error} If the instance is Nothing.
   */
  FromJust() {
    return this.IsNothing() && l.FromJustIsNothing(), this.value;
  }
  /**
   * Retrieves the value of a Just instance or returns a default value if it's Nothing.
   * @param defaultValue - The default value to return if the instance is Nothing.
   * @returns {T} The value of the instance if it's Just, otherwise the default value.
   */
  FromMaybe(e) {
    return this.hasValue ? this.value : e;
  }
  /**
   * Compares two Maybe instances for equality.
   * @param other - Another Maybe instance to compare with.
   * @returns {boolean} True if both instances are equal, otherwise false.
   */
  equals(e) {
    return this.IsJust() === e.IsJust() && (!this.IsJust() || this.FromJust() === e.FromJust());
  }
  /**
   * Compares two Maybe instances for inequality.
   * @param other - Another Maybe instance to compare with.
   * @returns {boolean} True if both instances are not equal, otherwise false.
   */
  notEquals(e) {
    return !this.equals(e);
  }
}
const r = o.Nothing, a = o.Just;
class f {
  /**
   * Creates an instance of LRUCache.
   * @param {CacheOptions} options - The options for the cache.
   */
  constructor(e) {
    this.capacity = e.maxSize, this.cache = /* @__PURE__ */ new Map(), this.head = r(), this.tail = r(), this.size = 0, this.maxAge = e.maxAge || Number.POSITIVE_INFINITY, this.onEviction = e.onEviction;
  }
  /**
   * Moves a node to the head of the LRU list.
   * @private
   * @param {LRUNode<K, V>} node - The node to move.
   */
  moveToHead(e) {
    const t = a(e);
    this.head.equals(t) || (e.prev.IsJust() && (e.prev.FromJust().next = e.next), e.next.IsJust() && (e.next.FromJust().prev = e.prev), this.tail.equals(t) && (this.tail = e.prev), this.head.IsJust() && (this.head.FromJust().prev = t), e.next = this.head, e.prev = r(), this.head = t, this.tail.IsNothing() && (this.tail = this.head));
  }
  /**
   * Sets a value in the cache for the specified key.
   * @param {K} key - The key to set.
   * @param {V} value - The value to set.
   * @param {{ maxAge?: number }} [options={}] - Options for the entry.
   * @returns {this} The cache instance.
   */
  set(e, t, { maxAge: i = this.maxAge } = {}) {
    const h = i !== Number.POSITIVE_INFINITY ? a(Date.now() + i) : r();
    let s = this.cache.get(e);
    if (s)
      s.value = t, this.moveToHead(s);
    else {
      s = {
        key: e,
        value: t,
        prev: r(),
        next: this.head,
        expiry: h
      };
      const c = a(s);
      if (this.head.IsJust() && (this.head.FromJust().prev = c), this.head = c, this.tail.IsNothing() && (this.tail = this.head), this.cache.set(e, s), this.size++, this.size > this.capacity) {
        const n = this.tail;
        this.cache.delete(n.ToChecked().key), n.ToChecked().prev && (n.ToChecked().prev.FromJust().next = r()), this.tail = n.ToChecked().prev, this.size--, this.onEviction && this.onEviction(n.ToChecked().key, n.ToChecked().value);
      }
    }
    return this;
  }
  /**
   * Deletes a value from the cache if it is expired.
   * @private
   * @param {K} key - The key to check.
   * @param {LRUNode<K, V>} item - The cache node to check.
   * @returns {boolean} True if the item was deleted, otherwise false.
   */
  deleteIfExpired(e, t) {
    return t.expiry.FromMaybe(Number.MAX_SAFE_INTEGER) <= Date.now() ? (this.delete(e), !0) : !1;
  }
  /**
   * Gets a value from the cache if it is not expired, otherwise deletes it.
   * @private
   * @param {K} key - The key to get.
   * @param {LRUNode<K, V>} item - The cache node to get.
   * @returns {Maybe<V>} The value if found and not expired, otherwise Nothing.
   */
  getOrDeleteIfExpired(e, t) {
    return this.deleteIfExpired(e, t) ? r() : (this.moveToHead(t), a(t.value));
  }
  /**
   * Peeks at a value in the cache without updating its position.
   * @param {K} key - The key to peek at.
   * @returns {Maybe<V>} The value if found, otherwise Nothing.
   */
  peek(e) {
    const t = this.cache.get(e);
    return t ? this.getOrDeleteIfExpired(e, t) : r();
  }
  /**
   * Gets a value from the cache for the specified key.
   * @param {K} key - The key to get.
   * @returns {Maybe<V>} The value if found, otherwise Nothing.
   */
  get(e) {
    const t = this.cache.get(e);
    return t ? this.getOrDeleteIfExpired(e, t) : r();
  }
  /**
   * Checks if the cache has the specified key.
   * @param {K} key - The key to check.
   * @returns {boolean} True if the key exists, otherwise false.
   */
  has(e) {
    return this.cache.has(e);
  }
  /**
   * Deletes the value from the cache for the specified key.
   * @param {K} key - The key to delete.
   * @returns {boolean} True if the key was deleted, otherwise false.
   */
  delete(e) {
    const t = this.cache.get(e);
    return t ? (t.prev.IsJust() ? t.prev.FromJust().next = t.next : this.head = t.next, t.next.IsJust() ? t.next.FromJust().prev = t.prev : this.tail = t.prev, this.cache.delete(e), this.size--, !0) : !1;
  }
  /**
   * Clears all values from the cache.
   */
  clear() {
    this.cache.clear(), this.head = r(), this.tail = r(), this.size = 0;
  }
  /**
   * Gets an iterable iterator of entries in the cache in ascending order.
   * @private
   * @returns {IterableIterator<[K, LRUNode<K, V>]>} An iterator of entries.
   */
  *entriesAscendingInternal() {
    for (const [e, t] of this.cache)
      this.deleteIfExpired(e, t) || (yield [e, t]);
  }
  /**
   * Emits eviction events for a set of cache entries.
   * @private
   * @param {Map<K, LRUNode<K, V>>} cache - The cache entries to evict.
   */
  emitEvictions(e) {
    if (typeof this.onEviction == "function")
      for (const [t, i] of e)
        this.onEviction(t, i.value);
  }
  /**
   * Resizes the cache to the new size.
   * @param {number} newSize - The new size of the cache.
   * @throws {TypeError} If the new size is not a positive number.
   */
  resize(e) {
    if (typeof e != "number" || e <= 0)
      throw new TypeError("`maxSize` must be a number greater than 0");
    const t = [...this.entriesAscendingInternal()], i = t.length - e;
    i > 0 ? (this.emitEvictions(new Map(t.slice(0, i))), this.cache = new Map(t.slice(i)), this.size = e) : (this.cache = new Map(t), this.size = t.length);
  }
  /**
   * Gets an iterable iterator of keys in the cache.
   * @returns {IterableIterator<K>} An iterator of keys.
   */
  *keys() {
    for (const [e] of this.entriesAscendingInternal())
      yield e;
  }
  /**
   * Gets an iterable iterator of values in the cache.
   * @returns {IterableIterator<V>} An iterator of values.
   */
  *values() {
    for (const [, e] of this.entriesAscendingInternal())
      yield e.value;
  }
  /**
   * Gets an iterable iterator of entries (key-value pairs) in the cache.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  *[Symbol.iterator]() {
    for (const [e, t] of this.entriesAscendingInternal())
      yield [e, t.value];
  }
  /**
   * Gets an iterable iterator of entries in the cache in descending order.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  *entriesDescending() {
    const e = [...this.cache].reverse();
    for (const [t, i] of e)
      this.deleteIfExpired(t, i) || (yield [t, i.value]);
  }
  /**
   * Gets an iterable iterator of entries in the cache in ascending order.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  *entriesAscending() {
    for (const [e, t] of this.entriesAscendingInternal())
      yield [e, t.value];
  }
  /**
   * Gets the maximum size of the cache.
   * @readonly
   * @type {number}
   */
  get maxSizeValue() {
    return this.capacity;
  }
  /**
   * Gets an iterable iterator of entries in the cache.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  entries() {
    return this.entriesAscending();
  }
  /**
   * Executes a callback function for each entry in the cache.
   * @param {(value: V, key: K, map: Cache<K, V>) => void} callback - The callback function to execute.
   * @param {any} [thisArg=this] - Value to use as this when executing the callback.
   */
  forEach(e, t = this) {
    for (const [i, h] of this.entriesAscending())
      e.call(t, h, i, this);
  }
  /**
   * Gets the string tag representation of the cache.
   * @readonly
   * @type {string}
   */
  get [Symbol.toStringTag]() {
    return JSON.stringify([...this.entriesAscending()]);
  }
}
class d {
  /**
   * Creates an instance of LFUCache.
   * @param {CacheOptions} options - The options for the cache.
   */
  constructor(e) {
    this.capacity = e.maxSize, this.cache = /* @__PURE__ */ new Map(), this.freqMap = /* @__PURE__ */ new Map(), this.minFreq = 0, this.size = 0, this.maxAge = e.maxAge || Number.POSITIVE_INFINITY, this.onEviction = e.onEviction;
  }
  /**
   * Updates the frequency of a node.
   * @private
   * @param {LFUNode<K, V>} node - The node to update.
   */
  updateFrequency(e) {
    var i, h, s;
    const t = e.freq;
    (i = this.freqMap.get(t)) == null || i.delete(e.key), ((h = this.freqMap.get(t)) == null ? void 0 : h.size) === 0 && (this.freqMap.delete(t), this.minFreq === t && this.minFreq++), e.freq++, this.freqMap.has(e.freq) || this.freqMap.set(e.freq, /* @__PURE__ */ new Set()), (s = this.freqMap.get(e.freq)) == null || s.add(e.key);
  }
  /**
   * Sets a value in the cache for the specified key.
   * @param {K} key - The key to set.
   * @param {V} value - The value to set.
   * @param {{ maxAge?: number }} [options={}] - Options for the entry.
   * @returns {this} The cache instance.
   */
  set(e, t, { maxAge: i = this.maxAge } = {}) {
    var c;
    const h = i !== Number.POSITIVE_INFINITY ? a(Date.now() + i) : r();
    let s = this.cache.get(e);
    return s ? (s.value = t, s.expiry = h, this.updateFrequency(s)) : (this.size >= this.capacity && this.evict(), s = {
      key: e,
      value: t,
      freq: 1,
      expiry: h
    }, this.cache.set(e, s), this.freqMap.has(1) || this.freqMap.set(1, /* @__PURE__ */ new Set()), (c = this.freqMap.get(1)) == null || c.add(e), this.minFreq = 1, this.size++), this;
  }
  /**
   * Evicts the least frequently used node from the cache.
   * @private
   */
  evict() {
    const e = this.freqMap.get(this.minFreq);
    if (e) {
      const t = e.values().next().value;
      e.delete(t), e.size === 0 && this.freqMap.delete(this.minFreq);
      const i = this.cache.get(t);
      i && (this.cache.delete(t), this.size--, this.onEviction && this.onEviction(t, i.value));
    }
  }
  /**
   * Deletes a node from the cache if it has expired.
   * @private
   * @param {K} key - The key to check.
   * @param {LFUNode<K, V>} item - The node to check.
   * @returns {boolean} True if the node was deleted, otherwise false.
   */
  deleteIfExpired(e, t) {
    return t.expiry.FromMaybe(Number.MAX_SAFE_INTEGER) <= Date.now() ? (this.delete(e), !0) : !1;
  }
  /**
   * Gets a node from the cache if it has not expired, otherwise deletes it.
   * @private
   * @param {K} key - The key to get.
   * @param {LFUNode<K, V>} item - The node to get.
   * @returns {Maybe<V>} The value if found, otherwise Nothing.
   */
  getOrDeleteIfExpired(e, t) {
    return this.deleteIfExpired(e, t) ? r() : (this.updateFrequency(t), a(t.value));
  }
  /**
   * Peeks at the value in the cache for the specified key without updating its frequency.
   * @param {K} key - The key to peek.
   * @returns {Maybe<V>} The value if found, otherwise Nothing.
   */
  peek(e) {
    const t = this.cache.get(e);
    return t ? this.getOrDeleteIfExpired(e, t) : r();
  }
  /**
   * Gets the value in the cache for the specified key.
   * @param {K} key - The key to get.
   * @returns {Maybe<V>} The value if found, otherwise Nothing.
   */
  get(e) {
    const t = this.cache.get(e);
    return t ? this.getOrDeleteIfExpired(e, t) : r();
  }
  /**
   * Checks if the cache has the specified key.
   * @param {K} key - The key to check.
   * @returns {boolean} True if the key exists, otherwise false.
   */
  has(e) {
    return this.cache.has(e);
  }
  /**
   * Deletes the value from the cache for the specified key.
   * @param {K} key - The key to delete.
   * @returns {boolean} True if the key was deleted, otherwise false.
   */
  delete(e) {
    var i, h;
    const t = this.cache.get(e);
    if (t) {
      const s = t.freq;
      return (i = this.freqMap.get(s)) == null || i.delete(e), ((h = this.freqMap.get(s)) == null ? void 0 : h.size) === 0 && (this.freqMap.delete(s), this.minFreq === s && this.minFreq++), this.cache.delete(e), this.size--, !0;
    }
    return !1;
  }
  /**
   * Clears all values from the cache.
   */
  clear() {
    this.cache.clear(), this.freqMap.clear(), this.minFreq = 0, this.size = 0;
  }
  /**
   * Gets an iterable iterator of entries in the cache in ascending order.
   * @private
   * @returns {IterableIterator<[K, LFUNode<K, V>]>} An iterator of entries.
   */
  *entriesAscendingInternal() {
    for (const [e, t] of this.cache)
      this.deleteIfExpired(e, t) || (yield [e, t]);
  }
  /**
   * Emits eviction events for a set of cache entries.
   * @private
   * @param {Map<K, LFUNode<K, V>>} cache - The cache entries to evict.
   */
  emitEvictions(e) {
    if (typeof this.onEviction == "function")
      for (const [t, i] of e)
        this.onEviction(t, i.value);
  }
  /**
   * Resizes the cache to the new size.
   * @param {number} newSize - The new size of the cache.
   * @throws {TypeError} If the new size is not a positive number.
   */
  resize(e) {
    if (typeof e != "number" || e <= 0)
      throw new TypeError("`maxSize` must be a number greater than 0");
    const t = [...this.entriesAscendingInternal()], i = t.length - e;
    i > 0 ? (this.emitEvictions(new Map(t.slice(0, i))), this.cache = new Map(t.slice(i)), this.size = e) : (this.cache = new Map(t), this.size = t.length);
  }
  /**
   * Gets an iterable iterator of keys in the cache.
   * @returns {IterableIterator<K>} An iterator of keys.
   */
  *keys() {
    for (const [e] of this.entriesAscendingInternal())
      yield e;
  }
  /**
   * Gets an iterable iterator of values in the cache.
   * @returns {IterableIterator<V>} An iterator of values.
   */
  *values() {
    for (const [, e] of this.entriesAscendingInternal())
      yield e.value;
  }
  /**
   * Gets an iterable iterator of entries (key-value pairs) in the cache.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  *[Symbol.iterator]() {
    for (const [e, t] of this.entriesAscendingInternal())
      yield [e, t.value];
  }
  /**
   * Gets an iterable iterator of entries in the cache in ascending order.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  *entriesAscending() {
    for (const [e, t] of this.entriesAscendingInternal())
      yield [e, t.value];
  }
  /**
   * Gets the maximum size of the cache.
   * @readonly
   * @type {number}
   */
  get maxSizeValue() {
    return this.capacity;
  }
  /**
   * Gets an iterable iterator of entries in the cache.
   * @returns {IterableIterator<[K, V]>} An iterator of entries.
   */
  entries() {
    return this.entriesAscending();
  }
  /**
   * Executes a callback function for each entry in the cache.
   * @param {(value: V, key: K, map: Cache<K, V>) => void} callback - The callback function to execute.
   * @param {any} [thisArg=this] - Value to use as this when executing the callback.
   */
  forEach(e, t = this) {
    for (const [i, h] of this.entriesAscending())
      e.call(t, h, i, this);
  }
  /**
   * Gets the string tag representation of the cache.
   * @readonly
   * @type {string}
   */
  get [Symbol.toStringTag]() {
    return JSON.stringify([...this.entriesAscending()]);
  }
}
export {
  d as LFUCache,
  f as LRUCache
};
